
#include "CACHE.hpp"

CACHE* CACHE::_instance=NULL;

CACHE* CACHE::getInstance() {
	if ( _instance == NULL ) {
		const unsigned long int l = 256*256;
		const unsigned long int s = 1024*4096;
		const unsigned long int k = 16;
		return ( _instance = new CACHE( l, s, k ) );
	} else
		return _instance;
	}

void CACHE::createInstance( unsigned long int l, unsigned long int s, unsigned long int k ) {
	if ( _instance == NULL )
		_instance = new CACHE( l, s, k );
	else
		std::cerr << "Warning: createInstance ignored since an instance was already created." << std::endl;
}

void CACHE::recreateInstance( unsigned long int l, unsigned long int s, unsigned long int k ) {
	if ( _instance == NULL )
		createInstance( l, s, k );
	else {
		delete _instance;
		_instance = NULL;
		createInstance( l, s, k );
	}
}

CACHE::CACHE( unsigned long int l, unsigned long int s, unsigned long int k ) {
	if ( ( s % l ) != 0 ) {
		std::cerr << "The size must be perfectly dividable by the number of cache lines!" << std::endl;
		exit( !EXIT_SUCCESS );
	}
	if ( ( l % k ) != 0 ) {
		std::cerr << "The total number of cache lines must be perfectly dividable between the number of sub caches!" << std::endl;
		std::cerr << "l % k = " << l << " % " << k << " = " << ( l % k ) << std::endl;
		exit( !EXIT_SUCCESS );
	}
	if ( k > l ) {
		std::cerr << "Number of subcaches cannot exceed total number of cache lines!" << std::endl;
		exit( !EXIT_SUCCESS );
	}
	_s = s; _l = l; _k = k;
	_sdl = s / l;
	_ldk = l / k;

#ifdef _TRACE
	std::cout << "---Cache line size set to " << _sdl << std::endl;
	std::cout << "---Number of cache lines per subcache is " << _ldk << std::endl;
#endif
	
	//gcc 4 thing
	typedef CACHE_SET_COLLECTION* CACHE_SET_COLLECTION_POINTER;
	//typedef DIRECT_MAPPED_CACHE* DIRECT_MAPPED_CACHE_POINTER;
	_csccollections = new CACHE_SET_COLLECTION_POINTER[ _ldk ];
	//_dmcaches = new DIRECT_MAPPED_CACHE_POINTER[ k ];
	
	//clear cache
	for( unsigned long int i=0; i<_ldk; i++ ) {
		_csccollections[ i ] = new CACHE_SET_COLLECTION( _k );
		//_dmcaches[ i ] = new DIRECT_MAPPED_CACHE( _ldk );
	}

	//clear status
	for( unsigned long int i=0; i<_ldk; i++ )
		_csccollections[ i ]->clear();
	REINIT = 0;
}

std::string CACHE::properties() {
	std::ostringstream oss;
	oss << "Cache size: " << _s << " bytes, number of cache lines: " << _l << ", number of set-associativity: " << _k << "," << std::endl;
	oss << "Cache line size: " << _sdl << " bytes, subcache size: " << _s / _k << " bytes." << std::endl;
	return oss.str();
}

void CACHE::clear() {
	for( unsigned long int i=0; i<_ldk; i++ )
		_csccollections[ i ]->clear();
	REINIT++;
}

void CACHE::clearStatistics() {
        _status._lc = 0; _status._bc = 0;
        _status._lm = 0; _status._bm = 0;
        _status._misses = 0; _status._initialMisses = 0;
        _status._hits = 0;
}

const STATUS &CACHE::getStatus() {
		return _status;
}

/* outdated!!
CACHE_LINE* CACHE::getLine( void* pointer, unsigned long int size, void* nextPointer, unsigned long int *size_left ) {
        for( unsigned long int i=0; i<_k; i++ ) {
		CACHE_LINE* line = &( _dmcaches[ i ]->getLine( pointer ) );
       		const unsigned long int line_start = (unsigned long int) line->_pointer;
		const unsigned long int pntr = (unsigned long int) pointer;
		const unsigned long int line_end = line_start + _sdl;
                if ( line_start <= pntr )
                        if ( pntr <= ( line_start + _sdl ) ) { // we have a hit
				if ( ( pntr + size ) > line_end ) { //line overflow
					nextPointer = (void*)( line_end + 1 );
					*size_left = size - (line_end - pntr);
				} else
					nextPointer = 0;
                       		return line;
			}
        }
        return 0;
}*/

CACHE_SET_COLLECTION* CACHE::getCacheLines( unsigned long int setID ) {
	return _csccollections[ setID ];
}

bool CACHE::hitOrMiss( void* pointer, unsigned long int size ) {
	unsigned long int memoryLinePointer = ( (unsigned long int)pointer ) / _sdl;
	unsigned long int setID = memoryLinePointer % _ldk;

	CACHE_SET_COLLECTION* curCol = getCacheLines( setID );
	bool result = curCol->bringIntoCache( memoryLinePointer );
	if ( size > _sdl ) {
		memoryLinePointer += _sdl;
		return result && hitOrMiss( (void*) memoryLinePointer, size-_sdl );
	} else
		return result;
}

/* outdated!!
bool CACHE::hitOrMiss( void* pointer, unsigned long int size ) {
	void *next;
	unsigned long int size_left;
	if ( getLine( pointer, size, next, &size_left ) == 0 )
		return false;
	else {
		if ( next == 0 )
			return true;
		else
			return hitOrMiss( next, &size_left );
	}
}*/

void CACHE::access( void* pointer, const unsigned long int size, unsigned long int &FAC ) {
	bool obligatory = false;
        unsigned long int memoryLinePointer = ( (unsigned long int)pointer ) / _sdl;
	unsigned long int memoryLineAddress = memoryLinePointer * _sdl;
        unsigned long int setID = memoryLinePointer % _ldk;
	unsigned long int diff  = (unsigned long int)pointer - memoryLineAddress;

#ifdef _TRACE
	std::cout << "Pointer " << (unsigned long int)pointer << " corresponds to memory line " << memoryLinePointer << std::endl;
	std::cout << "which in turn corresponds to the set ID " << setID << "/" << _ldk << std::endl;
#endif

        CACHE_SET_COLLECTION* curCol = getCacheLines( setID );
        const bool result = curCol->bringIntoCache( memoryLinePointer );
	//unsigned long int used_bandwith = ...

	//there is always cache to cpu bandwith used:
	//_status->bc += used_bandwith;

#ifdef _DEBUG
	std::cout << pointer << " ";
#endif
	if ( result ) {
		_status._hits++;
#ifdef _DEBUG
		std::cout << " HIT" << std::endl;
#endif
	} else {
		//at least a partial miss; a complete cache line is wapped in
		//_status->bm += _sdl;

		//we count hits and misses per line
		if( FAC <= REINIT ) {
			_status._initialMisses++;
			obligatory = true;
		} else
			_status._misses++;
#ifdef _DEBUG
		std::cout << " MISS" << std::endl;
#endif
	}

        if ( size > _sdl ) {
		//requested data spans multiple cache lines
#ifdef _DEBUG
		std::cout << "..spans multiple lines" << std::endl;
#endif
                memoryLinePointer += _sdl;
                hitOrMiss( (void*) ( memoryLineAddress + _sdl ) , size - _sdl + diff );
	} else
		if( obligatory )
			FAC = REINIT + 1;
			
}

/*void CACHE::read( void* pointer, unsigned long int size ) {
		void *next;
		unsigned long int size_left;
		if ( getLine( pointer, size, next, &size_left ) == 0 ) {
				//total cache miss
				_status->misses++;
				swapIn( pointer, size );
				return false;
		} else {
				if ( next == 0 ) {
						//cache miss, complete status update
						_status->hits++;
						_status->lc++;
						_status->bc += size;
						return true;
				} else {
						//at least a partial hit, only update bandwith
						_status->bc += size;
						//is everything was in cache, _hits and _lc will be updated
						//correctly in the final recusive call to CACHE::read()
						
						//if there is only a partial hit, _misses will be updated correctly.
						//swapIn will be called which accounts for which parts after the missing
						//part are also cached so to update _lc, _bc, _lm and _bm correctly.
						return read( next, &size_left );
				}
		}
}*/

/*void CACHE::write( void* pointer, unsigned long int size ) {
		//TODO (keep in mind: write-back cache)
}

void CACHE::swapIn( void* pointer, unsigned long int size ) {
		//TODO (keep in mind: write-back cache)
}*/

